/* JSGalBuilder: (c) 2009 by Andreas S. Krebs
 * 
 * This file is part of JSGalBuilder.
 *
 * JSGalBuilder is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * JSGalBuilder is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JSGalBuilder.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sberk.jsgal.gui;

import java.awt.Component;
import java.awt.Desktop;
import java.awt.Rectangle;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;
import java.lang.reflect.Array;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.net.URI;
import java.util.ArrayList;

import javax.swing.SwingUtilities;
import javax.swing.UIManager;
import javax.xml.parsers.ParserConfigurationException;

import net.sberk.util.MiscUtil;
import net.sberk.util.XMLUtil;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

@SuppressWarnings("nls")
class BuilderPrefs
{
	private final static int	NUM_RECENT_PROJECTS=5;
	private final static String	TAG_ROOTEL="JSGALBUILDER_PREFS";
	private final static String	ATTR_VERSION="VERSION";
	private final static String	TAG_PARAM="PARAM";
	private final static String	ATTR_NAME="NAME";
	private final static String	ATTR_VALUE="VALUE";
	
	private final static String	PREF_FILE=System.getProperty("user.home")+File.separator+".jsgb.prefs";

	@PersistInConfig public static String	LOOK_AND_FEEL=UIManager.getSystemLookAndFeelClassName();
	@PersistInConfig public static boolean	USE_SYSTEM_BROWSER=true;
	@PersistInConfig public static String	BROWSER_PATH="";
	@PersistInConfig public static ImageHelper.GenerationOptions	GEN_OPTS=new ImageHelper.GenerationOptions();
	@PersistInConfig public static int			MAIN_SPLIT_DIV_LOC=200;
	@PersistInConfig public static Rectangle	MAIN_WINDOW_RECT=new Rectangle(-1, -1, 1024, 768);
	@PersistInConfig public static String		RECENT_PROJECTS[]=new String[0];
	
	public static void startBrowser(URI uri) throws IOException
	{
		if(USE_SYSTEM_BROWSER)
		{
			if(uri.toString().startsWith("mailto"))
				Desktop.getDesktop().mail(uri);
			else
				Desktop.getDesktop().browse(uri);
		}
		else
			Runtime.getRuntime().exec(new String[] { BROWSER_PATH, uri.toString() });
	}
	
	public static void addRecentProject(File f)
	{
		if(f!=null)
		{
			String	path=f.getAbsolutePath();
			
			ArrayList<String> al=new ArrayList<String>();
			
			al.add(path);
			int	i=1;
			for(String s: RECENT_PROJECTS)
			{
				if(!MiscUtil.equals(path, s))
				{
					al.add(s);
					
					i++;
					if(i>=NUM_RECENT_PROJECTS)
						break;
				}
			}
			
			RECENT_PROJECTS=al.toArray(new String[al.size()]);
		}
	}
	
	public static File[] getRecentProjects()
	{
		File	fa[]=new File[RECENT_PROJECTS.length];
		
		for(int i=0; i<RECENT_PROJECTS.length; i++)
			fa[i]=new File(RECENT_PROJECTS[i]);
		
		return fa;
	}
	
	public static void setLookAndFeel(Component c, String lfClassName) throws Exception
	{
		UIManager.setLookAndFeel(lfClassName);
		SwingUtilities.updateComponentTreeUI(c);
		
		// finally update the L&F property
		LOOK_AND_FEEL=lfClassName;
	}
	
	public static void savePrefs()
	{
		try
		{
			Document	doc=toDOM();
			
			XMLUtil.writeXML(doc, new FileOutputStream(PREF_FILE));
		}
		catch(Exception e)
		{
			// since we're about to exit, we simply dump the stack
			e.printStackTrace();
		}
	}
	
	public static void loadPrefs(Component c)
	{
		try
		{
			File	f=new File(PREF_FILE);
			
			if(f.isFile() && f.canRead())
			{
				Document	doc=XMLUtil.parseXML(f, false);
				
				parseDOM(doc);
			}
		}
		catch(Exception e)
		{
			// we simply dump the stack, don't bother the user
			e.printStackTrace();
		}
		
		// finally set the look and feel
		try
		{
			setLookAndFeel(c, LOOK_AND_FEEL);
		}
		catch(Exception e)
		{
			// we simply dump the stack, don't bother the user
			e.printStackTrace();
		}
	}
	
	private static Document toDOM() throws ParserConfigurationException, IllegalArgumentException, SecurityException, IllegalAccessException
	{
		Document doc=XMLUtil.createDOM();
		
		// create root element and add all fields
		Element	el=doc.createElement(TAG_ROOTEL);
		el.setAttribute(ATTR_VERSION, I18N.JSGB_VERSION);
		
		doc.appendChild(el);
		
		// all fields should be static, so null as instance is fine
		addFields(el, BuilderPrefs.class.getFields(), null);
		
		return doc;
	}
	
	private static void addFields(Element el, Field fs[], Object instance) throws IllegalArgumentException, IllegalAccessException
	{
		Document	doc=el.getOwnerDocument();
		
		for(Field f: fs)
		{
			Object	val=f.get(instance);
				
			if(val==null || Modifier.isFinal(f.getModifiers()))	
				continue;
			
			Element	pEl=doc.createElement(TAG_PARAM);
			pEl.setAttribute(ATTR_NAME, f.getName());
			
			Class<?>	type=f.getType();
			if(type.isPrimitive() || type.equals(String.class) || type.isEnum()) 
			{
				pEl.setAttribute(ATTR_VALUE, val.toString());
				el.appendChild(pEl);
			}
			else if(type.isArray())
			{
				Class<?> ctype=type.getComponentType();
				if(ctype.isPrimitive() || ctype.equals(String.class) || ctype.isEnum()) 
				{
					el.appendChild(pEl);
					
					for(int i=0; i<Array.getLength(val); i++)
					{
						Element	aEl=doc.createElement(TAG_PARAM);
						
						aEl.setAttribute(ATTR_VALUE, Array.get(val, i).toString());
						pEl.appendChild(aEl);
					}
				}
			}
			else if(f.isAnnotationPresent(PersistInConfig.class))
			{
				// this is a complex object
				Field	ofs[]=type.getFields();
				
				if(ofs.length>0)
				{
					el.appendChild(pEl);
					
					addFields(pEl, ofs, val);
				}
			}
		}
	}

	private static void parseDOM(Document doc) throws IllegalArgumentException, SecurityException, IllegalAccessException
	{
		Element	el=doc.getDocumentElement();
		
		if(el.getTagName().equals(TAG_ROOTEL))
		{
			getFields(el, BuilderPrefs.class.getFields(), null);
		}
	}
	
	private static void getFields(Element el, Field fs[], Object instance) throws IllegalArgumentException, IllegalAccessException
	{
		NodeList	nl=el.getChildNodes();
		for(int i=0; i<nl.getLength(); i++)
		{
			if(!(nl.item(i) instanceof Element))
				continue;
			
			Element ce=(Element)nl.item(i);
			String	name=ce.getAttribute(ATTR_NAME);
			
			if(!ce.getTagName().equals(TAG_PARAM) || name==null || name.length()<=0)
				continue;
			
			// find matching field
			for(Field f: fs)
			{
				if(name.equals(f.getName()))
				{
					Class<?>	type=f.getType();
					if(type.isPrimitive() || type.equals(String.class) || type.isEnum())
					{
						f.set(instance, getPrimitiveValue(type, ce.getAttribute(ATTR_VALUE), f.get(instance)));
					}
					else if(type.isArray())
					{
						Class<?> ctype=type.getComponentType();
						if(ctype.isPrimitive() || ctype.equals(String.class) || ctype.isEnum()) 
						{
							ArrayList<Object>	al=new ArrayList<Object>();
							
							NodeList	nl2=ce.getChildNodes();
							for(int j=0; j<nl2.getLength(); j++)
							{
								if(!(nl2.item(j) instanceof Element))
									continue;
								
								Element	ve=(Element)nl2.item(j);
								
								al.add(getPrimitiveValue(ctype, ve.getAttribute(ATTR_VALUE), null));
							}
							
							f.set(instance, al.toArray((Object[])Array.newInstance(ctype, al.size())));
						}
					}
					else if(f.isAnnotationPresent(PersistInConfig.class))
					{
						Object	obj=f.get(instance);
						
						getFields(ce, type.getFields(), obj);
					}
					
					break;
				}
			}
		}
	}
	
	@SuppressWarnings("unchecked")
	private static Object getPrimitiveValue(Class<?> type, String val, Object defVal)
	{
		try
		{
			if(type.equals(String.class))
				return val;
			else if(type.equals(byte.class))
				return new Byte(val);
			else if(type.equals(short.class))
				return new Short(val);
			else if(type.equals(int.class))
				return new Integer(val);
			else if(type.equals(long.class))
				return new Long(val);
			else if(type.equals(float.class))
				return new Float(val);
			else if(type.equals(double.class))
				return new Double(val);
			else if(type.equals(char.class))
				return new Character(val.charAt(0));
			else if(type.equals(boolean.class))
				return new Boolean(val);
			else if(type.isEnum())
				return Enum.valueOf((Class<? extends Enum>)type, val);
			else
				throw new Exception("no conversion from String to "+type.getName());
		}
		catch(Exception e)
		{
			// we simply ignore ivalid values
			e.printStackTrace();
		}
		
		return defVal;
	}
	
	
	// define the annotation for fields to store (only for needed for not primitive objects)
	@Retention(RetentionPolicy.RUNTIME)
	@Target(ElementType.FIELD)
	@interface PersistInConfig { /* empty */ }
}